/*
 * Copyright (C) 2017-2021 Dominique Dhoutaut, Thierry Arrabal, Eugen Dedu.
 *
 * This file is part of BitSimulator.
 *
 * BitSimulator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BitSimulator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BitSimulator.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "routing-agent.h"


//==============================================================================
//
//          RoutingAgent  (class)
//
//==============================================================================

int RoutingAgent::aliveAgents = 0;
int RoutingAgent::forwardedDataPackets = 0;
int RoutingAgent::anchor1id, RoutingAgent::anchor2id, RoutingAgent::anchor3id; 

RoutingAgent::RoutingAgent(Node *_hostNode): hostNode(_hostNode) {
  aliveAgents++;
}

RoutingAgent::~RoutingAgent() {
  //     aliveAgents--;
}

void RoutingAgent::receivePacketFromApplication(PacketPtr _packet) {
  _packet->setSrcSequenceNumber(hostNode->getNextSrcSequenceNumber());
  hostNode->enqueueOutgoingPacket(_packet);
}


void RoutingAgent::processAckTimedOut(PacketPtr /* _packet */) {
  cout << " Routing Agent processAckTimedOut" << endl; /*getchar();*/
}

void RoutingAgent::processBackoffSending(int /* packetId */)
{
  cout << " Routing Agent processBackoffSending" << endl;
}

void RoutingAgent::processSLRBackoffSending(int /* packetId */)
{
  cout << " Routing Agent processSLRBackoffSending" << endl;
}

void RoutingAgent::processDominantForwarder(int /* packetId */)
{
  cout << " Routing Agent processDominantForwarder" << endl;
}


///////  helpers for backoff flooding-based agents

simulationTime_t RoutingAgent::getBackoffDynamic (Node *hostNode, PacketPtr packet, float multiplier, mt19937_64 *rng) {
  // the backoff window = 2 times sending each message pulse, plus the
  // maximum propagation time, cf. Backoff flooding article

  simulationTime_t backoffWindow =
    2 * ((Node::getPulseDuration() * (1 + packet->beta * (packet->size - 1)))
    + ScenarioParameters::getCommunicationRange() / PROPAGATIONSPEED);

  if (hostNode->getEstimatedNeighbours() != -1)
    backoffWindow *= hostNode->getEstimatedNeighbours(); //deden
  else
    backoffWindow *= hostNode->getNeighboursCount(); //nodeden

  backoffWindow *= multiplier;

  uniform_int_distribution<unsigned long> distrib (0, backoffWindow);

  return distrib (*rng);
}

///////  helpers for backoff ring flooding

simulationTime_t RoutingAgent::getBackoffDynamicRing (Node *hostNode, PacketPtr packet, float multiplier, mt19937_64 *rng) {
  // the backoff window = 2 times sending each message pulse, plus the
  // maximum propagation time, cf. Backoff flooding article

  simulationTime_t backoffWindow =
    2 * ((Node::getPulseDuration() * (1 + packet->beta * (packet->size - 1)))
    + ScenarioParameters::getCommunicationRange() / PROPAGATIONSPEED);

  if (hostNode->getEstimatedNeighbours() != -1)
    backoffWindow *= (hostNode->getEstimatedNeighbours ()) * (hostNode->getRingNeighboursCount(ScenarioParameters::getRangeSmall(),ScenarioParameters::getRangeBig()))/(hostNode->getNeighboursCount());
  else
    backoffWindow *= hostNode->getRingNeighboursCount(ScenarioParameters::getRangeSmall(),ScenarioParameters::getRangeBig());

  backoffWindow *= multiplier;

  uniform_int_distribution<unsigned long> distrib (0, backoffWindow);
  return distrib (*rng);
}


////////  helpers for slr-based agents

void RoutingAgent::slrOpenFile (ofstream &file) {
  string filename = ScenarioParameters::getScenarioDirectory() + "/" + ScenarioParameters::getPrefix() + "SLRPositions" + ScenarioParameters::getDefaultExtension();
  file.open(filename);
  if (!file) {
    cerr << "*** ERROR *** While opening SLR positions file " << filename << endl;
    exit(EXIT_FAILURE);
  }
}

void RoutingAgent::readCommRangeSetupAttribute () {
  long cr;
  tinyxml2::XMLElement *slrElement = ScenarioParameters::getXMLRootNode()->FirstChildElement("routing")->FirstChildElement(ScenarioParameters::getRoutingAgentName().c_str());
  ScenarioParameters::queryLongAttr (slrElement, "commRangeSetup_nm", "commRangeSetup_mm", cr, false, nullptr, ScenarioParameters::getCommunicationRange(), "no attribute");
  if (cr > ScenarioParameters::getCommunicationRange()) {
    cerr << "*** ERROR *** commRangeSetup must be <= to commRange " << endl;
    exit(EXIT_FAILURE);
  }
  ScenarioParameters::setCommunicationRangeSetup (cr);
}

void RoutingAgent::readAnchorAttributes () {
  tinyxml2::XMLElement *slrElement = ScenarioParameters::getXMLRootNode()->FirstChildElement("routing")->FirstChildElement(ScenarioParameters::getRoutingAgentName().c_str());

  ScenarioParameters::queryIntAttr (slrElement, "anchor1id", anchor1id, true, nullptr, 0, "no anchor1id attribute");
  ScenarioParameters::queryIntAttr (slrElement, "anchor2id", anchor2id, true, nullptr, 0, "no anchor2id attribute");
  ScenarioParameters::queryIntAttr (slrElement, "anchor3id", anchor3id, false, nullptr, -1, "no anchor3id attribute");  // TODO make it mandatory in 3D case?
}

////////  helpers for ring-based agents

void RoutingAgent::readRangeAttributes () {
  long rb, rs;
  tinyxml2::XMLElement *ringElement = ScenarioParameters::getXMLRootNode()->FirstChildElement("routing")->FirstChildElement(ScenarioParameters::getRoutingAgentName().c_str());
  rb = ScenarioParameters::queryLongAttribute (ringElement, "rangeBig_nm", "rangeBig_mm");
  rs = ScenarioParameters::queryLongAttribute (ringElement, "rangeSmall_nm", "rangeSmall_mm");
  if (ScenarioParameters::getRangeBig() == -1)
    ScenarioParameters::setRangeBig (rb);
  if (ScenarioParameters::getRangeSmall() == -1)
    ScenarioParameters::setRangeSmall (rs);
  if (ScenarioParameters::getRangeBig() < ScenarioParameters::getRangeSmall()) {
    cerr << "*** ERROR *** rangeBig must be greater that rangeSmall" << endl;
    exit(EXIT_FAILURE);
  }
}

////////  helpers for proba-based agents

void RoutingAgent::readProbaAttribute () {
  tinyxml2::XMLElement *probaElement = ScenarioParameters::getXMLRootNode()->FirstChildElement("routing")->FirstChildElement(ScenarioParameters::getRoutingAgentName().c_str());
  float proba = ScenarioParameters::queryFloatAttribute (probaElement, "probability");
  if (ScenarioParameters::getProbability() == 0)
    ScenarioParameters::setProbability (proba);
}
